import typing
from datetime import datetime
from pprint import pprint

import bson

from .init_creds import init_mongo, test
from .dates import set_next_vigency, calculate_payment_date, get_periodo, localize, DIAS_ANTICIPO_DE_PAGO, \
    today_argentina
from .format import datetime_parser
from .mercadopago_ops import get_payments_from_user_id
from dateutil.relativedelta import *
import mercadopago
from bson import ObjectId
from muvisdk.SDK import SDK


def get_client_price(client: dict, plan_price: int, corpo_discount: float, sum_access=True,
                     access_count_limit: int = 9999) -> dict:
    """ Obtener el precio a pagar

    :param client: cliente con la estructura que se usa en mongodb
    :type client: dict
    :param plan_price: precio del plan nativo
    :type plan_price: int
    :param corpo_discount: porcentaje de descuento del plan corporativo
    :type corpo_discount: float
    :param sum_access: Si es True, suma los accesos desde el último cobro
    :type sum_access: bool
    :param access_count_limit: tope de cantidad de accesos pagos a cobrar
    :type access_count_limit: int
    :return: informacion del precio
    :rtype: dict
    """
    prices = {
        "native_plan": plan_price,
        "descuentos_absolutos": [],
        "descuentos_porcentuales": [],
        "descuentos_masivos": []
    }

    if "discounts" in client.keys():
        if client["discounts"]:
            try:
                for discount in [i for i in client["discounts"] if i["aplicaciones_restantes"] > 0]:
                    concepto = discount.get("concepto", "n/a")
                    if not concepto:
                        concepto = "n/a"

                    if discount.get("codigo") == "descuento masivo" and discount["monto_absoluto"] > 0:
                        discount_element = {
                            "concepto": concepto,
                            "descuento": round(discount["monto_absoluto"], 2)
                        }
                        prices["descuentos_masivos"].append(discount_element)
                        plan_price = plan_price - discount["monto_absoluto"]
                        continue

                    if discount["monto_absoluto"] > 0:
                        discount_element = {
                            "concepto": concepto,
                            "descuento": round(discount["monto_absoluto"], 2)
                        }
                        prices["descuentos_absolutos"].append(discount_element)
                        # plan_price = plan_price - discount["monto_absoluto"]

                    if discount["porcentaje"] > 0:
                        discount_element = {
                            "concepto": concepto,
                            "descuento": discount["porcentaje"]
                        }
                        prices["descuentos_porcentuales"].append(discount_element)
            except:
                print(f"    ERROR en el cálculo de descuentos - se retorna un diccionario de precios vacío.")
                return {}

    if corpo_discount:
        if corpo_discount > 100 or corpo_discount < 0:
            print(f"    ERROR en el cálculo de descuentos - se retorna un diccionario de precios vacío.")
            return {}
        discount_element = {
            "concepto": "corpo_discount",
            "descuento": corpo_discount
        }
        prices["descuentos_porcentuales"].append(discount_element)

    subtotal = round(plan_price * (100 - sum([d["descuento"] for d in prices["descuentos_porcentuales"]])) / 100, 2)
    subtotal = max(0, subtotal)

    if sum_access:
        accesos = get_access_amount(client["_id"], max_count=access_count_limit)
        prices["access"] = accesos
    else:
        accesos = 0

    prices["final_price"] = round(subtotal + accesos, 2)
    if prices["final_price"] < 1:
        prices["final_price"] = 0  # mercadopago no permite pagos de menos de $1

    charges_detail = {
        'final_price': prices['final_price'],
        'native_plan': prices['native_plan']
    }
    if accesos > 0:
        charges_detail['access'] = accesos

    for d in prices['descuentos_masivos']:
        if d['concepto'] not in charges_detail and "{} 1".format(d['concepto']) not in charges_detail:
            concepto = d['concepto']
        else:
            concepto = "{} 1".format(d['concepto'])
            if d['concepto'] in charges_detail:
                charges_detail[concepto] = charges_detail[d['concepto']]
                del charges_detail[d['concepto']]
            concepto = "{} {}".format(d['concepto'], len([k for k in charges_detail if d['concepto'] in k]) + 1)

        charges_detail[concepto] = - d['descuento']

    for d in prices['descuentos_porcentuales']:
        if d['concepto'] not in charges_detail and "{} 1".format(d['concepto']) not in charges_detail:
            concepto = d['concepto']
        else:
            concepto = "{} 1".format(d['concepto'])
            if d['concepto'] in charges_detail:
                charges_detail[concepto] = charges_detail[d['concepto']]
                del charges_detail[d['concepto']]
            concepto = "{} {}".format(d['concepto'], len([k for k in charges_detail if d['concepto'] in k]) + 1)

        charges_detail[concepto] = round(- plan_price * d['descuento'] / 100, 2)

    for d in prices['descuentos_absolutos']:
        if d['concepto'] not in charges_detail and "{} 1".format(d['concepto']) not in charges_detail:
            concepto = d['concepto']
        else:
            concepto = "{} 1".format(d['concepto'])
            if d['concepto'] in charges_detail:
                charges_detail[concepto] = charges_detail[d['concepto']]
                del charges_detail[d['concepto']]
            concepto = "{} {}".format(d['concepto'], len([k for k in charges_detail if d['concepto'] in k]) + 1)

        charges_detail[concepto] = - d['descuento']
        charges_detail['final_price'] -= d['descuento']

    print(f"    El monto a cobrar es de ${charges_detail['final_price']}")
    return charges_detail


def get_access_amount(c_id: bson.ObjectId, max_count: int = 9999) -> int:
    """ Obtener monto de acceso

    :param c_id: id del socio
    :type c_id: bson.ObjectId
    :param max_count: Límite de cantidad de accesos con costo a cobrar.
    :type max_count: int
    :return: Monto de acceso
    :rtype: int
    """
    db = init_mongo()
    last_payment = [p["original_payment_date"] for p in db.boletas.find({"member_id": c_id})]
    if last_payment:
        last_payment_date = max(last_payment)
    else:
        last_payment_date = datetime(year=2000, month=1, day=1)
    uses = db.accesos.find({"socio_id": c_id, "date": {"$gt": last_payment_date}, "costo": {"$gt": 0}}).limit(max_count)
    result = sum([a["costo"] for a in uses])
    print(f"    El monto de accesos a cobrar es de {result} acumulados desde {last_payment_date}.")
    return result


def substract_discounts(client: dict):
    """ Descontar aplicaciones de descuento

    :param client: cliente con la estructura que se usa en mongodb
    :type client: dict
    """
    if client["discounts"]:
        for discount in client["discounts"]:
            if discount["aplicaciones_restantes"] > 0:
                discount["aplicaciones_restantes"] -= 1
                print("    se descuenta una aplicacion restante de descuento.")


############
# PAYMENTS #
############

def calculate_poi(client, merchant_id):
    point_of_interaction = {
        "transaction_data": {}
    }

    if "poi" in client.keys() and client['poi']:
        first_time_use = False
        number = client["poi"]["installments"] + 1
        subscription_sequence = str(client["poi"]["payment_reference"])
        point_of_interaction["transaction_data"]["payment_reference"] = {"id": subscription_sequence}
    else:
        first_time_use = True
        number = 1

    db = init_mongo()
    merchant = db.merchants.find_one(merchant_id)

    point_of_interaction["transaction_data"].update({
        "first_time_use": first_time_use,
        "subscription_sequence": {
            "number": number,
            "total": 32
        },
        "invoice_period": {
            "period": number,
            "type": "monthly"
        },
        "billing_date": today_argentina().strftime("%Y-%m-%d"),
        "subscription_id": f"{merchant['credentials']['mercadopago']['application_id']}-{client['documento']}"
    })
    point_of_interaction["type"] = "SUBSCRIPTIONS"

    return point_of_interaction


def resolve_payment(payment_data: dict,
                    card_data: dict,
                    _client: dict,
                    _plan: dict,
                    sdk: typing.Union[SDK, None],
                    sdk_cadena: SDK = None) -> dict:
    """ Resolver pago si es posible

    :param payment_data: informacion del pago
    :type payment_data: dict
    :param card_data: informacion de la tarjeta
    :type card_data: dict
    :param _client: informacion del cliente
    :type _client: dict
    :param _plan: informacion del plan
    :type _plan: dict
    :param sdk: sdk del negocio
    :type sdk: SDK
    :param sdk_cadena: sdk del merchant cadena
    :type sdk: SDK
    :return: informacion con el resultado del proceso
    :rtype: dict
    """
    if not payment_data["charges_detail"]:
        print("    Error al calcular el precio del cliente.")
        return {
            "status": "error",
            "status_detail": "Error al calcular el precio del cliente.",
            "date_created": today_argentina()
        }

    if payment_data["charges_detail"]["final_price"] == 0:
        print("    Pago no realizado por ser precio 0.")
        return {
            "status": "approved",
            "status_detail": "Pago no realizado por ser precio 0.",
            "processor": None,
            'payment_type': None,
            "id": 11,
            "date_created": today_argentina()
        }

    if _client["preferred_payment_method"] == "efectivo":
        print("    Boleta generada con status pending_efectivo")
        return {
            "status": "pending_efectivo"
        }

    if "id" not in card_data or not card_data['id']:
        print("    Tarjeta no encontrada. No se puede procesar el pago.")
        return {
            "status": "error",
            "status_detail": "No se puede obtener la tarjeta activa.",
            "date_created": today_argentina()
        }

    if not sdk:
        print("    El merchant no tiene ningun prcesador habilitado")
        return {
            "status": "error",
            "status_detail": "no_payment_processor",
            "date_created": today_argentina(),
            "processor": None
        }

    print("    Se procede a realizar un pago de $" + str(payment_data["charges_detail"]["final_price"]))
    payment_result = process_payment(payment_data, card_data, _client, _plan, sdk, sdk_cadena)
    return payment_result


def process_payment(payment_data: dict,
                    card_data: dict,
                    _client: dict,
                    _plan: dict,
                    sdk: SDK,
                    sdk_cadena: SDK = None) -> dict:
    """ Procesar pago

    :param payment_data: informacion del pago
    :type payment_data: dict
    :param card_data: informacion de la tarjeta
    :type card_data: dict
    :param _client: informacion del cliente
    :type _client: dict
    :param _plan: informacion
    :type _plan: dict
    :param sdk: sdk del negocio
    :type sdk: SDK
    :param sdk: sdk del merchant cadena
    :type sdk_cadena: SDK
    :return: informacion con el resultado del proceso
    :rtype: dict
    """
    if not sdk_cadena or 'application_fee' not in payment_data:
        sdk_cadena = sdk

    print("    Se procede a realizar el pago con MP")
    # data = {"card_id": int(card_data["id"])}
    card_token_response = sdk_cadena.card_token().create(card_data)

    if card_token_response["status"] > 299:
        print("         Falló la creación de token de tarjeta.")
        print(card_token_response["response"])
        return {
            "status": "error",
            "status_detail": "Falló la creación de token de la tarjeta.",
            "response": card_token_response["response"]["message"],
            "date_created": today_argentina(),
            "processor": card_token_response["response"]["processor"]
        }

    card_token = card_token_response["response"]["id"]
    print(f"        Token generado OK")

    if "tries" in payment_data.keys():
        n_try = len(payment_data["tries"]) + 1
        numero_intento = "Reintento " + str(n_try)
    else:
        numero_intento = "Primer intento"

    try:
        street_number = int(_client["domicilio"]["altura"])
    except:
        street_number = None
    mp_payment_data = {
        "additional_info": {
            "items": [{
                "id": str(_plan['_id']),
                "title": f'{_plan["name"]} - {_client["nombre"]} {_client["apellido"]} - {_client["email"]} -'
                         f'{payment_data["period"]}',
                "category_id": _plan["nivel_de_acceso"],
                "description": numero_intento + "- id: " + f'{_client["_id"]}'
            }],
            "shipments": {
                "receiver_address": {
                    "state_name": _client["domicilio"]["provincia"],
                    "city_name": _client["domicilio"]["localidad"],
                    "street_name": _client["domicilio"]["calle"],
                    "street_number": street_number,
                }
            },
            "payer": {
                "first_name": _client["nombre"],
                "last_name": _client["apellido"],
                "address": None
            }
        },
        "notification_url": None if test else f"https://apisportclub.xyz/notificacion/mp"
                                              f"?source_news=webhooks&merchant={str(_plan['merchant_id'])}",
        "transaction_amount": round(payment_data["charges_detail"]["final_price"], 2),
        "token": card_token,  # ??
        "description": _plan["name"],  # ir a base de datos de planes y traer el name
        "bin": card_data['first_six_digits'] if 'first_six_digits' in card_data else None,
        "installments": 1,  # ??
        "payer": _client,
        "external_reference": 'Buen Club - {} - {}'.format(payment_data['source'], _client['documento'])
    }

    merchant_id = payment_data['merchant_id']

    if "payment_method_id" in card_data:
        mp_payment_data["payment_method_id"] = card_data["payment_method_id"]

    if "application_fee" in payment_data.keys():
        if payment_data['charges_detail']['final_price'] > payment_data["application_fee"] > 0:
            merchant_id = ObjectId('6178652dcfb117ad58a2cd3d')
            mp_payment_data["application_fee"] = payment_data["application_fee"]
    
    if payment_data['status'] != 'restored':
        mp_payment_data["point_of_interaction"] = calculate_poi(_client, merchant_id)

    payment_attempt = sdk.payment().create(mp_payment_data)
    payment_response = datetime_parser(payment_attempt["response"])

    print("         El estado del pago es:", str(payment_response["status"]))
    if "id" in payment_response:
        print("         El id del pago:", str(payment_response["id"]))

    # si hay algun error en el pago
    if payment_attempt["status"] >= 299:
        print(payment_response)
        try:
            detail = payment_response["message"]
        except KeyError:
            detail = "Falló el intento de pago."
        return {
            "status": "error",
            "status_detail": detail,
            "response": payment_attempt["response"],
            "date_created": today_argentina(),
            "processor": payment_attempt["response"]["processor"]
        }

    # si el pago es 200
    return {
        "status": payment_response["status"],
        "status_detail": payment_response["status_detail"],
        "id": payment_response["id"],
        "date_created": today_argentina(),
        "processor": payment_attempt["response"]["processor"]
    }


def create_payment_data(_client: dict,
                        prices: dict,
                        merchant_id: ObjectId,
                        source: str,
                        seller: dict = None,
                        insert=True,
                        id_operacion=None) -> dict:
    """ Crear data del pago y guardalo en la coleccion de boletas de mongodb

    :param _client: cliente con la estructura que se usa en mongodb
    :type _client: dict
    :param prices: informacion de precios
    :type prices: dict
    :param merchant_id: id del merchant
    :type merchant_id: ObjectId
    :param source: de dónde proviene la boleta - checkout o recurring_charges
    :type source: str
    :param insert: si es True, se inserta boleta en base de datos
    :type source: bool
    :return: data que se guardo en la coleccion de boletas
    :rtype: dict
    """
    db = init_mongo()

    if "checkout" in source or "cobro_aon" in source:
        opd = today_argentina()
        period = get_periodo(opd)
    else:
        opd = _client["next_payment_date"]
        period = get_periodo(opd + relativedelta(days=DIAS_ANTICIPO_DE_PAGO))

    data = {
        "member_id": _client["_id"],
        "date_created": today_argentina(),
        "original_payment_date": opd,
        "source": source,
        "tries": [],
        "status": "error_not_processed",
        "merchant_id": merchant_id,
        "charges_detail": prices,
        "period": period,
        "plan_id": _client["active_plan_id"]
    }
    if seller:
        data["application_fee"] = seller["fee"]
        data["seller_merchant_id"] = seller["merchant_id"]

    if 'cobro_aon' in source and id_operacion is None:
        data['id_operacion'] = id_operacion

    if insert:
        _id = db.boletas.insert_one(data)
        print(f"    Se inserta una nueva boleta de pago con _id: {_id.inserted_id}")

    return data


def update_payment_data(payment_result: dict, payment_data: dict, card_data: dict):
    """ Actualizar pago. En payment_data se agrega un nuevo intento de pago

    :param payment_result: Resultado del pago
    :type payment_result: dict
    :param payment_data: data del pago que se quiere actualizar
    :type payment_data: dict
    :param card_data: informacion de la tarjeta
    :type card_data: dict
    """
    n_try = len(payment_data["tries"]) + 1

    if "id" not in payment_result.keys():
        payment_result["id"] = 400  # corresponde a un error 400

    intento = {
        "try_number": n_try,
        "payment_day": payment_result["date_created"],
        "payment_type": card_data["card_type"],
        "card_brand": card_data["card_brand"],
        "card_id": card_data["id"],
        "status": payment_result["status"],
        "status_detail": payment_result["status_detail"],
        "payment_id": payment_result["id"],
        "processor": payment_result["processor"] if "processor" in payment_result else None
    }

    payment_data["tries"].append(intento)
    if payment_data["status"] != 'restored' or payment_result["status"] == 'approved':
        payment_data["status"] = payment_result["status"]


def format_card(mp_card):
    if set(mp_card.keys()) == {'id', 'payer_id', 'card_type', 'card_brand', 'last_four_digits', 'issuer',
                               'cardholder', 'expiration_month', 'expiration_year'}:
        return mp_card
    try:
        card = {
            'id': mp_card['id'],
            'card_type': mp_card['payment_method']['payment_type_id'],
            'card_brand': mp_card["payment_method"]["name"],
            'last_four_digits': mp_card['last_four_digits'],
            'issuer': {
                'name': mp_card['issuer']['name']
            },
            'cardholder': {
                'name': mp_card['cardholder']['name']
            },
            'expiration_month': mp_card['expiration_month'],
            'expiration_year': mp_card['expiration_year'],
            'payer_id': mp_card['user_id']
        }
        if 'valid' in mp_card:
            card['valid'] = mp_card['valid']
        return card
    except KeyError:
        return mp_card


def get_empty_card():
    return {
        "id": None,
        "payer_id": None,
        "card_type": None,
        "card_brand": None
    }


def get_cards(client: dict):
    cards = []
    if 'cards' not in client:
        return cards

    active_card = client['active_card'] if 'active_card' in client else None

    for card in client['cards']:
        if 'status' in card and card['status'] == 'invalid':
            continue
        formatted_card = format_card(card)
        if formatted_card['id'] == active_card:
            cards.insert(0, formatted_card)
        else:
            cards.append(formatted_card)

    return cards


def get_active_card(_client: dict, price: dict, sdk: mercadopago.SDK):
    """ Obtener la tarjeta activa (con la que se realizan los pagos)

    :param _client: Cliente con la estructura que se usa en mongodb
    :type _client: dict
    :param price: Precio que se quiere cobrar. Si es 0 no se busca la tarjeta
    :type price: dict
    :param sdk: sdk de mercado pago
    :type sdk: mercadopago.SDK
    :return: data de la tarjeta
    :rtype: dict
    """

    card = get_empty_card()
    if not price:
        print("    No existe precio - se devuelve tarjeta vacía")
        return card

    if price["final_price"] == 0:
        print("         Por monto $0 no se busca tarjeta.")
        return card

    if "cards" in _client.keys():
        if _client["cards"]:
            try:
                card_complete = next((item for item in _client["cards"] if item["id"] == _client["active_card"]), card)
                if card_complete["id"]:
                    card = card_complete
                else:
                    card = _client["cards"][0]
                print("    Se obtiene tarjeta correctamente.")
                return {"id": card["id"],
                        "payer_id": card["customer_id"],
                        "card_type": card["payment_method"]["payment_type_id"],
                        "card_brand": card["payment_method"]["name"]}
            except:
                print("         ERROR: No se puede obtener la tarjeta del cliente.")
                return card
        else:
            print("         Cliente sin atributo 'cards'.")

    if not card["id"] and "active_card" in _client.keys() and "mercadopago_id" in _client.keys():
        mp_card_data = sdk.card().get(_client["mercadopago_id"], _client["active_card"])
        if mp_card_data["status"] == 200:
            mp_card_data = mp_card_data["response"]
            card["id"] = mp_card_data["id"]
            card["payer_id"] = mp_card_data["user_id"]
            card["card_type"] = mp_card_data["payment_method"]["payment_type_id"]
            card["card_brand"] = mp_card_data["payment_method"]["name"]
            print("         Se reconstruyó tarjeta a partir de MP exitosamente.")
        elif "active_card" in _client.keys():
            card["id"] = _client["active_card"]

    return card


def update_client(_client: dict, _payment_result: dict, _plan: dict, period: str, update_npd: bool = False):
    """ Actualizar cliente

    :param _client: Cliente con la estructura que se usa en mongodb
    :type _client: dict
    :param _payment_result: Data del resultado del pago
    :type _payment_result: dict
    :param _plan: Plan con la estructura que se usa en mongodb
    :type _plan: dict
    :param period: periodo de la última boleta generada
    :type period: str
    :param update_npd: indica si se debe actualizar el next_payment_date del cliente
    :type update_npd: bool
    """
    # si el pago tiene id_mp se los agrego al cliente
    if "id" in _payment_result.keys():
        print("         Se inserta payment id en el cliente.")
        _client["payment_ids"].insert(0, _payment_result["id"])
        _client["last_payment_id"] = _payment_result["id"]

    # actualizaciones de cliente
    _client["lastModified"] = today_argentina()

    if update_npd:
        _client["next_payment_date"] = calculate_payment_date(_client["period_init_day"], _plan["cobro"], period)
        print("         El próximo dia de cobro es el " + _client["next_payment_date"].strftime("%d/%m/%y"))
        # se resta una aplicacion de descuento
        if "discounts" in _client.keys():
            substract_discounts(_client)

    # updateo de la fecha de vigencia y los cobros recurrentes
    if _payment_result["status"] == "approved":
        _client["cobros_recurrentes"] = _client["cobros_recurrentes"] + 1 if "cobros_recurrentes" in _client else 1
        if _client["next_payment_date"]:
            _client["fecha_vigencia"] = set_next_vigency(_client["next_payment_date"])
            print("         Se actualiza la fecha de vigencia: {}".format(_client["fecha_vigencia"]))

        pago_decidir = 'processor' in _payment_result and _payment_result['processor'] == 'decidir'
        if len(str(_payment_result["id"])) > 4 and not pago_decidir:  # es un pago de mercadopago
            if "poi" in _client.keys():
                _client["poi"]["installments"] += 1
            else:
                _client["poi"] = {"installments": 1, "payment_reference": _payment_result["id"]}

    else:
        if localize(_client["fecha_vigencia"]) < today_argentina():
            _client["status"] = "inactivo"
            print("         El cliente pasa inactivo por vencer la vigencia")
    return


def restore_pending_payment(user_id, boleta_id):
    """ Actualiza una boleta que quedó en estado 'pending' a su estado real en función del último pago en mercadopago

    :param user_id: id de mongo del usuario
    :type user_id: str
    :param boleta_id: id de mongo de la boleta en estado pendiente
    """
    db = init_mongo()
    boleta = db.boletas.find_one({"_id": ObjectId(boleta_id)})
    if boleta["status"] != "error_not_processed":
        print("Boleta no estaba rota")
        return
    client = db.clientes.find_one({"_id": ObjectId(user_id)})
    plan = db.planes.find_one({"_id": client["active_plan_id"]})
    merchant = db.merchants.find_one({"_id": plan["merchant_id"]})
    print(merchant["credentials"]["mercadopago"]["access_token"])
    sdk = mercadopago.SDK(merchant["credentials"]["mercadopago"]["access_token"])
    user_id = client["mercadopago_id"].split('-')[0]
    payment_response = get_payments_from_user_id(user_id, sdk, days=6, limit=1)[-1]

    payment_result = {
        "status": payment_response["status"],
        "status_detail": payment_response["status_detail"],
        "id": payment_response["id"],
        "date_created": payment_response["date_created"]
    }
    card = {"id": payment_response["card"]["id"],
            "payer_id": payment_response["card"]["cardholder"]["identification"]["number"],
            "card_type": payment_response["payment_type_id"],
            "card_brand": payment_response["payment_method_id"]}

    update_payment_data(payment_result, boleta, card)
    db.boletas.update_one({"_id": boleta["_id"]}, {"$set": boleta})

    update_client(client, payment_result, plan, boleta["period"])
    db.clientes.update_one({"_id": client["_id"]}, {"$set": client})


def refund_process(payment_id: str, amount: int, sdk: mercadopago.SDK) -> dict:
    """ Hacer un reembolso

    :param payment_id: id del pago a reembolsar
    :type payment_id: str
    :param amount: monto a reembolsar
    :type amount: int
    :param sdk: sdk de mercadopago
    :type sdk: mercadopago.SDK
    :return: respuesta de mercadopago con informacion del reembolso
    :rtype: dict
    """
    refund_data = {
        "amount": amount
    }
    refund_response = sdk.refund().create(payment_id, refund_data)
    refund = refund_response["response"]
    pprint(refund)
    refund_data = datetime_parser(refund)
    return refund_data


def add_try_to_pending_boleta(boleta: dict, mp_payment: dict) -> None:
    db = init_mongo()
    payment_result = build_payment_result(mp_payment)
    card = get_card_data_from_payment(mp_payment)
    update_payment_data(payment_result, boleta, card)
    db.boletas.update_one({"_id": boleta["_id"]}, {"$set": boleta})


def build_payment_result(mp_payment: dict) -> dict:
    pr = {
        "status": mp_payment["status"],
        "status_detail": mp_payment["status_detail"],
        "id": mp_payment["id"],
        "date_created": mp_payment["date_created"]
    }
    return datetime_parser(pr)


def get_card_data_from_payment(mp_payment: dict) -> dict:
    return {
        "id": mp_payment["card"]["id"],
        "payer_id": mp_payment["payer"]["id"],
        "card_type": mp_payment["payment_type_id"],
        "card_brand": mp_payment["payment_method_id"]
    }


def get_other_cards(_client, card_data):
    active_card_id = card_data["id"]
    cards_list = []
    for card in _client["cards"]:
        if "id" in card.keys():
            if card["id"]:
                if active_card_id == card["id"]:
                    continue
                else:
                    card_data = {
                        "card_id": card["id"],
                        "payer_id": card["customer_id"],
                        "card_type": card["payment_method"]["payment_type_id"],
                        "card_brand": card["payment_method"]["name"]
                    }
                    cards_list.append(card_data)
    return cards_list


def create_payment_data_for_sdk(price: int,
                                client: dict,
                                boleta: dict,
                                card_data: dict,
                                nro_de_intento: int,
                                installments: int = 1,
                                merchant: dict = None,           # Puede enviarse aparte, no dentro de client
                                plan: dict = None,                 # Puede enviarse aparte, no dentro de client
                                seller_fee: int = None) -> dict:
    # Si entran separado del client
    if merchant:
        client['merchant'] = merchant
    if plan:
        client['plan'] = plan
    # El client debe tener el plan y el merchant incluidos
    payment_data = dict()

    # DATOS DE LA TARJETA #################################
    # Usar la response de CardToken.create() del SDK

    if card_data['status'] == 'error':
        return {'status': 'error', 'status_detail': 'error_empty_card'}
    card_data = card_data['response']
    payment_data['token'] = card_data['id']
    if 'bin' in card_data:
        payment_data['bin'] = card_data['bin']
    elif 'first_six_digits' in card_data:
        payment_data['bin'] = card_data['first_six_digits']

    # Payer ###############################################

    payment_data['payer'] = {
        'nombre': client['nombre'],
        'apellido': client['apellido'],
        'email': client['email'],
        'mercadopago_id': client['mercadopago_id'] if 'mercadopago_id' in client else None,
        'decidir_id': client['decidir_id'] if 'decidir_id' in client else None,
    }

    # MP aditional info ###################################

    try:
        street_number = int(client["domicilio"]["altura"])
    except:
        street_number = None
    if nro_de_intento == 0:
        numero_intento = "Primer intento"
    else:
        numero_intento = "Reintento " + str(nro_de_intento)

    payment_data['additional_info'] = {
        'items': [{
            "id": str(client['plan']['_id']),
            "title": f"{client['plan']['name']} - {client['nombre']} {client['apellido']} - {client['email']} - {boleta['period']}",
            "category_id": client['plan']["nivel_de_acceso"],
            "description": numero_intento + " - id: " + f"{client['_id']}"
        }],
        'shipments': {
            "receiver_address": {
                "state_name": client["domicilio"]["provincia"],
                "city_name": client["domicilio"]["localidad"],
                "street_name": client["domicilio"]["calle"],
                "street_number": street_number}
        },
        'payer': {
            "first_name": client["nombre"],
            "last_name": client["apellido"],
            "address": None
        }
    }
    payment_data["notification_url"] = None if test else f"https://apisportclub.xyz/notificacion/mp?source_news=webhooks&merchant={str(client['merchant']['_id'])}",
    payment_data['description'] = client['plan']['name']
    payment_data['installments'] = installments # 1 por default
    if 'source' in boleta and 'documento' in client:
        payment_data['external_reference'] = 'Buen Club - ' + boleta['source'] + ' - ' + client['documento']
    else:
        payment_data['external_reference'] = 'Buen Club'

    # POINT OF INTERACTION ################################

    point_of_interaction = {
        "transaction_data": {}
    }

    if "poi" in client.keys():
        first_time_use = False
        number = client["poi"]["installments"] + 1
        subscription_sequence = str(client["poi"]["payment_reference"])
        point_of_interaction["transaction_data"]["payment_reference"] = {"id": subscription_sequence}
    else:
        first_time_use = True
        number = 1

    point_of_interaction["transaction_data"].update({
        "first_time_use": first_time_use,
        "subscription_sequence": {
            "number": number,
            "total": 32
        },
        "invoice_period": {
            "period": number,
            "type": "monthly"
        },
        "billing_date": today_argentina().strftime("%Y-%m-%d"),
        "subscription_id": f"{client['merchant']['credentials']['mercadopago']['application_id']}-{client['documento']}"
    })
    point_of_interaction["type"] = "SUBSCRIPTIONS"
    payment_data["point_of_interaction"] = point_of_interaction

    # Add Seller fee ######################################

    if seller_fee:
        if 100 > seller_fee > 0:
            payment_data["application_fee"] = round((1 - seller_fee/100) * price, 2)

    payment_data['transaction_amount'] = round(price, 2)

    return payment_data



import base64
import io
import json
from datetime import datetime
from pprint import pprint

import numpy as np
import qrcode
from babel.numbers import format_number
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.x509.oid import NameOID
from dateutil.relativedelta import relativedelta
from PyPDF2 import PdfFileReader, PdfFileWriter
from reportlab.lib.colors import HexColor
from reportlab.lib.pagesizes import A4
from reportlab.lib.units import mm
from reportlab.lib.utils import ImageReader
from reportlab.pdfgen import canvas
from zeep.helpers import serialize_object

from .dates import today_argentina
from .init_creds import init_mongo

TIPO_DOCUMENTO = {
    'CUIT': 80,
    'CUIL': 86,
    'CDI': 87,
    'LE': 89,
    'LC': 90,
    'CI_extranjera': 91,
    'acta_nacimiento': 93,
    'pasaporte': 94,
    'DNI': 96,
}

TIPO_COMPROBANTE = {
    'factura_a': 1,
    'nota_debito_a': 2,
    'nota_credito_a': 3,
    'factura_b': 6,
    'nota_debito_b': 7,
    'nota_credito_b': 8,
    'factura_c': 11,
    'nota_debito_c': 12,
    'nota_credito_c': 13
}

LETRA_FACTURA = {
    1: 'A',
    2: 'A',
    3: 'A',
    6: 'B',
    7: 'B',
    8: 'B',
    11: 'C',
    12: 'C',
    13: 'C',
}

TIPO_CONCEPTO = {
    'producto': 1,
    'servicios': 2,
    'protuctos_servicios': 3
}

TIPO_IVA = {
    21.0: 5,
    10.5: 4,
    0: 3
}


def generateKey(cuit: str, bits: int = 2048, save_file=False):
    """ Funcion para generar una private key con python

    :param cuit: CUIT a generar la PK
    :type cuit: int
    :param bits: Bits de la clave privada (2048 por default)
    :type bits: int
    :param type_pk: tipo de clave, por default TYPE_RSA
    :type type_pk: crypto Type
    :param save_file: si se desea guardar la clave en un archivo, False por default
    :type save_file: boolean
    :return: key
    :rtype: Key de PyOpenSSL

    """

    pk = rsa.generate_private_key(public_exponent=65537, key_size=bits, backend=default_backend)

    if save_file:
        keyfile = 'pk_' + cuit + '.key'
        f = open(keyfile, "wb")
        pk_byte = pk.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.TraditionalOpenSSL,
            encryption_algorithm=serialization.NoEncryption()
        )
        f.write(pk_byte)
        f.close()

    return pk


def generateCSR(cuit: str, razon_social: str, key, save_file=False):
    """ Funcion para generar el CSR

    :param cuit: CUIT a generar la PK
    :type cuit: int
    :param razon_social: Razon social del facturante
    :type razon_social: str
    :param save_file: si se desea guardar el certificado en un archivo, False por default
    :type save_file: boolean
    :return: CSR
    :rtype: string

    """

    CSR = x509.CertificateSigningRequestBuilder().subject_name(x509.Name([
        x509.NameAttribute(NameOID.COMMON_NAME, "muvi_facturacion"),
        x509.NameAttribute(NameOID.COUNTRY_NAME, "AR"),
        x509.NameAttribute(NameOID.ORGANIZATION_NAME, razon_social),
        x509.NameAttribute(NameOID.SERIAL_NUMBER, "CUIT" + cuit)
    ])).sign(key, hashes.SHA256())

    if save_file:
        csrfile = f'cert_{str(cuit)}.csr'
        f = open(csrfile, "wb")
        f.write(CSR.public_bytes(serialization.Encoding.PEM))
        f.close()
    return CSR


def format_numero_factura(pto_venta: int, cbte_num: int, tipo_cbte: int, short=False) -> str:
    if np.isnan(pto_venta) or np.isnan(cbte_num) or np.isnan(tipo_cbte):
        return '-'

    # Version corta
    if short:
        return "%03d" % (tipo_cbte,) + '_' + "%05d" % (pto_venta,) + '_' + "%08d" % (cbte_num,)

    letra = LETRA_FACTURA.get(tipo_cbte)
    # Version larga
    if tipo_cbte in [1, 6, 11]:
        nro_factura = f'Factura {letra} - N°' + "%05d" % (pto_venta,) + '-' + "%08d" % (cbte_num,)
    elif tipo_cbte in [3, 8, 13]:
        nro_factura = f'Nota de Crédito {letra} - N°' + "%05d" % (pto_venta,) + '-' + "%08d" % (cbte_num,)
    return nro_factura


def limit_string(string: str,limit: int) -> str:

    if len(string) > limit:
        return string[:limit]
    else:
        return string


def process_razon_social (rz: str, limit_long: str):
    first_row = ''
    second_row = ''
    if len(rz) > limit_long:
        rz_list = rz.split(' ')
        long_string = 0
        for word in rz_list:
            long_string += len(word)
            if long_string <= limit_long:
                first_row = first_row + word + ' '
            else:
                second_row = second_row + word + ' '
    else:
        first_row = rz
    return [first_row, second_row]


def make_qr(factura: dict, qr_path: str):

    url = "https://www.afip.gob.ar/fe/qr/?p=%s"

    try:
        documento = int(factura['data_cliente']['documento'])
    except TypeError:
        documento = factura['data_cliente']['documento']

    fecha_factura = factura['data_factura']['fecha_factura'].strftime("%Y-%m-%d")

    tipo_doc = {
        1: 80,
        2: 80,
        3: 80
    }

    data = {
        'ver': 1,
        'fecha': fecha_factura,
        'cuit': int(factura['data_facturante']['cuit']),
        'ptoVta': int(factura['data_factura']['pto_venta']),
        'tipoCmp': int(factura['data_factura']['tipo_comprobante']),
        'nroCmp': int(factura['data_factura']['num_comprobante']),
        'importe': float(factura['data_venta']['importe_total']),
        'moneda': 'PES',
        'ctz': float(1.000),
        'tipoDocRec': tipo_doc.get(factura['data_factura']['tipo_comprobante'], 96),
        'nroDocRec': documento,
        'tipoCodAut': 'E',
        'codAut': int(factura['data_cae']['cae'])
    }

    data_json = json.dumps(data)
    url = url % (base64.b64encode(data_json.encode('ascii')).decode('ascii'))

    qr = qrcode.QRCode(
        version=1,
        error_correction=qrcode.constants.ERROR_CORRECT_L,
        box_size=10,
        border=4
    )
    qr.add_data(url)
    qr.make(fit=True)

    img = qr.make_image(fill_color='black', back_color='white')

    img.save(qr_path, "PNG")


def make_factura_pdf(files_names: dict, factura: dict, brand: dict = {}) -> str:
    """
    :param files_names: Nombres de los archivos,
    FILES_NAMES = {
        'QR_PATH':'...',
        'TEMPLATE_FACTURA':'...',
        'TEMPLATE_NOTA_DE_CREDITO':'...',
        'PDF_NAME':'...'}
    :param factura: Dict de factura proveniente de mongo
    :return: Lista con los archivos que genero
    """
    packet = io.BytesIO()
    can = canvas.Canvas(packet)
    x_page = 210
    y_page = 297
    can.setPageSize(A4)

    can.setFillColorRGB(1, 1, 1)
    can.rect(0, (y_page - 20) * mm, width=230, height=60, fill=1, stroke=0)

    if (logo_facturacion := brand.get('images', {}).get('logo_facturacion')):
        logo_facturacion = ImageReader(logo_facturacion)
        can.drawImage(logo_facturacion, (x_page - 215) * mm, (y_page - 42) * mm, width=235, height=170, mask='auto')

    if (ribbon := brand.get('color_palette', {}).get('facturacion_ribbon', '#cccccc')):
        can.setFillColor(HexColor(ribbon))
        can.rect(0, (y_page - 27) * mm, width=210 * mm, height=10, fill=1, stroke=0)

    #### FORMAT DEL PRECIO ################################
    importe_neto = '$ ' + format_number(factura['data_venta']['importe_neto'], locale='es')
    importe_iva = '$ ' + format_number(factura['data_venta']['importe_iva'], locale='es')
    importe_total = '$ ' + format_number(factura['data_venta']['importe_total'], locale='es')

    #### TIPO DE BOLETA ###################################
    can.setFillColorRGB(0.25, 0.25, 0.25)
    can.setFont("Helvetica-Bold", 42)
    can.drawString(99.5 * mm, (y_page - 45) * mm, LETRA_FACTURA.get(factura['data_factura']['tipo_comprobante']))
    can.setFont("Helvetica-Bold", 12)
    can.drawString(96.5 * mm, (y_page - 52) * mm, f"COD. 0{factura['data_factura']['tipo_comprobante']}")

    #### DATOS DEL FACTURANTE #############################
    can.setFillColorRGB(1, 1, 1)
    can.setFont("Helvetica-Bold", 11)
    numero_factura = format_numero_factura(
        int(factura['data_factura']['pto_venta']),
        int(factura['data_factura']['num_comprobante']),
        int(factura['data_factura']['tipo_comprobante'])
    )
    can.drawString(126 * mm, (y_page - 35.2) * mm, numero_factura)

    can.setFillColorRGB(0, 0, 0)
    can.setFont("Helvetica", 10)
    can.drawString(126 * mm, (y_page - 41.5) * mm, 'Fecha: ' + factura['data_factura']['fecha_factura'].strftime("%Y-%m-%d"))
    cuit = factura['data_facturante']['cuit']
    can.drawString(165 * mm, (y_page - 41.5) * mm, f'CUIT: {cuit[0:2]}-{cuit[2:-1]}-{cuit[-1]}')
    razon_social = process_razon_social(factura['data_facturante']['razon_social'], 25)
    can.drawString(126 * mm, (y_page - 47) * mm, 'Razon Social: ' + razon_social[0])
    can.drawString(126 * mm, (y_page - 52.5) * mm, razon_social[1])
    can.drawString(126 * mm, (y_page - 58) * mm, 'Inicio de Actividades: '+factura['data_facturante']['inicio_de_actividades'].strftime("%Y-%m-%d"))
    can.drawString(126 * mm, (y_page - 63.5) * mm, 'Ingresos Brutos: '+ factura['data_facturante']['ii_bb'])

    can.setFillColorRGB(0.2, 0.2, 0.2)
    can.setFont("Helvetica-Bold", 11)
    can.drawString(8 * mm, (y_page - 36) * mm, 'DATOS DEL FACTURANTE')
    can.setFillColorRGB(0, 0, 0)
    can.setFont("Helvetica", 10)
    direccion_facturante = limit_string(factura['data_facturante']['domicilio']['calle'],45) + ' ' + factura['data_facturante']['domicilio']['altura']
    can.drawString(8 * mm, (y_page - 41.5) * mm, 'Dirección: '+direccion_facturante)
    can.drawString(8 * mm, (y_page - 47) * mm, limit_string(factura['data_facturante']['domicilio']['localidad'],50))
    can.drawString(8 * mm, (y_page - 52.5) * mm, f"Teléfono: {factura['data_facturante']['telefono']}")
    can.drawString(8 * mm, (y_page - 58) * mm, f"Email: {factura['data_facturante']['email']}")
    can.drawString(8 * mm, (y_page - 63.5) * mm, 'IVA RESPONSABLE INSCRIPTO')

    #### DATOS DE LAS CONDICIONES DE VENTA ################
    can.setFillColorRGB(0.2, 0.2, 0.2)
    can.setFont("Helvetica-Bold", 10)
    can.drawString(120 * mm, (y_page - 74) * mm, 'CONDICIONES DE VENTA')
    can.setFillColorRGB(0, 0, 0)
    can.setFont("Helvetica", 9)
    metodo_de_pago = factura['data_factura'].get('payment_type', '-') or '-'
    payment_types = {
        'cash': 'Efectivo',
        'debit_card': 'Débito',
        'credit_card': 'Crédito',
    }
    metodo_de_pago = payment_types.get(metodo_de_pago, '-')
    can.drawString(120 * mm, (y_page - 82) * mm, 'Método de pago: ' + metodo_de_pago)
    can.drawString(120 * mm, (y_page - 88) * mm, 'Tipo: Servicios')
    can.drawString(120 * mm, (y_page - 94) * mm, 'Fecha de inicio de servicios: ' + factura['data_venta']['inicio_servicios'].strftime("%Y-%m-%d"))
    can.drawString(120 * mm, (y_page - 100) * mm, 'Fecha de fin de servicios: ' + factura['data_venta']['fin_servicios'].strftime("%Y-%m-%d"))
    can.drawString(120 * mm, (y_page - 106) * mm, 'Fecha de pago del servicio: ' + factura['data_venta']['fecha_pago'].strftime("%Y-%m-%d"))

    #### DATOS DEL CLIENTE ################################
    can.setFillColorRGB(0.2, 0.2, 0.2)
    can.setFont("Helvetica-Bold", 10)
    can.drawString(8 * mm, (y_page - 74) * mm, 'INFORMACIÓN DEL CLIENTE')
    can.setFillColorRGB(0, 0, 0)
    can.setFont("Helvetica", 10)
    cliente_name = limit_string(factura['data_cliente']['nombre'],30) + " " + factura['data_cliente']['apellido']
    can.drawString(8 * mm, (y_page - 80) * mm, 'Cliente: '+limit_string(cliente_name,55))
    direccion_cliente = limit_string(factura['data_cliente']['domicilio']['calle'],45) + ' ' + factura['data_cliente']['domicilio']['altura']
    can.drawString(8 * mm, (y_page - 86) * mm, 'Direccion: '+ limit_string(direccion_cliente,55))
    can.drawString(8 * mm, (y_page - 92) * mm, 'Provincia: ' + factura['data_cliente']['domicilio']['provincia'])
    can.drawString(8 * mm, (y_page - 98) * mm, 'Email: '+factura['data_cliente']['email'])

    condicion = "Consumidor Final"
    tipo_doc = "Documento"
    letra = LETRA_FACTURA.get(factura['data_factura']['tipo_comprobante'])
    if letra == "A":
        condicion = "Responsable Inscripto"
        tipo_doc = "CUIT"

    can.drawString(8 * mm, (y_page - 104) * mm, f"{tipo_doc}: {factura['data_cliente']['documento']}")
    can.drawString(8 * mm, (y_page - 110) * mm, f'Condición: {condicion}')

    #### CONCEPTOS ########################################
    can.setFillColorRGB(0, 0, 0)
    can.setFont("Helvetica", 10)
    # Cantidad
    can.drawString(10 * mm, (y_page - 128.5) * mm, "1,00")
    # Descripcion del Plan
    can.drawString(42 * mm, (y_page - 128.5) * mm, factura['data_plan']['plan_name'])
    # Subtotal
    can.drawString(150 * mm, (y_page - 128.5) * mm, importe_neto)
    # Total
    can.drawString(178 * mm, (y_page - 128.5) * mm, importe_neto)
    # Si se desea agregar otro item deberia ser en la altura 138.5

    #### PRICES ###########################################
    can.setFillColorRGB(0, 0, 0)
    can.setFont("Helvetica",11)
    can.drawString(120 * mm, (y_page - 226) * mm, "Subtotal")
    can.drawString(165 * mm, (y_page - 226) * mm, importe_neto)
    can.drawString(120 * mm, (y_page - 240) * mm, "IVA 21%")
    can.drawString(165 * mm, (y_page - 240) * mm, importe_iva)
    can.setFillColorRGB(1, 1, 1)
    can.setFont("Helvetica-Bold", 15)
    can.drawString(120 * mm, (y_page-252) * mm, "Final Total")
    can.drawString(165 * mm, (y_page-252) * mm, importe_total)

    #### CAE y QR #########################################
    make_qr(factura, files_names['QR_PATH'])
    can.setFillColorRGB(0.1, 0.1, 0.1)
    can.setFont("Helvetica-Bold", 11)
    can.drawString(46 * mm, (y_page-268) * mm, 'N° de CAE: ' + factura['data_cae']['cae'])
    can.drawString(46 * mm, (y_page-278) * mm, 'Fecha de Vencimiento: ' + factura['data_cae']['fecha_vencimiento'].strftime('%d/%m/%Y'))
    can.drawInlineImage(files_names['QR_PATH'], x=15*mm, y=(y_page-284.5)*mm, width=26 * mm, height=26 * mm)

    #### NOTA DE CREDITO ##################################
    if 'data_nota_de_credito' in factura:
        can.setFillColorRGB(0, 0, 0)
        can.setFont("Helvetica", 10)
        letra = LETRA_FACTURA.get(factura['data_factura']['tipo_comprobante'])
        can.drawString(8 * mm, (y_page - 226) * mm, f'Nota de crédito {letra}, asociada con:')

        numero_factura = format_numero_factura(
            pto_venta=factura['data_nota_de_credito']['factura_asociada']['pto_venta'],
            cbte_num=factura['data_nota_de_credito']['factura_asociada']['num_comprobante'],
            tipo_cbte=factura['data_nota_de_credito']['factura_asociada']['tipo_comprobante']
        )
        can.drawString(8 * mm, (y_page - 232) * mm, numero_factura)

    #### SE GUARDA EL PDF #################################
    can.save()
    packet.seek(0)
    new_pdf = PdfFileReader(packet)
    if 'TEMPLATE_FOLDER' in files_names:
        template_factura = files_names['TEMPLATE_FOLDER'] + files_names['TEMPLATE_FACTURA']
    else:
        template_factura = files_names['TEMPLATE_FACTURA']

    existing_pdf = PdfFileReader(open(template_factura, "rb"))

    output = PdfFileWriter()
    page = existing_pdf.getPage(0)
    page.mergePage(new_pdf.getPage(0))
    output.addPage(page)
    if 'FOLDER_PDF' in files_names:
        file_name_pdf = files_names['FOLDER_PDF']+files_names['PDF_NAME']
    else:
        file_name_pdf = files_names['PDF_NAME']
    outputStream = open(file_name_pdf, "wb")
    output.write(outputStream)
    outputStream.close()
    return file_name_pdf


def make_name_pdf(factura: dict) -> str:
    numero_factura = format_numero_factura(
        int(factura['data_factura']['pto_venta']),
        int(factura['data_factura']['num_comprobante']),
        int(factura['data_factura']['tipo_comprobante']),
        short=True
    )
    name_pdf = f"{factura['data_facturante']['cuit']}_{numero_factura}.pdf"
    return name_pdf


def check_afip_data(merchant: dict, list_to_check: list) -> int:
    print('<--- CHECKS AFIP START --->')
    errors = 0
    if 'afip' in merchant:
        for check in list_to_check:
            if check in merchant['afip']:
                print(f"OK: {check}")
            else:
                errors += 1
                print(f"ERROR: No se encuentra {check}")
        if 'activate_facturacion' in merchant['afip'] and not merchant['afip']['activate_facturacion']:
            print("ERROR: El merchant no tiene activado el proceso de facturacion")
            errors += 1
    else:
        errors += 1
        print('ERROR: El merchant no tiene data de AFIP')
    print('<--- CHECKS AFIP END --->')
    return errors


def process_period(boleta: dict):
    pid = boleta['cliente']['period_init_day']
    month, year = boleta['period'].split('/')
    try:
        month = int(month)
        year = int(year)
    except ValueError:
        today = today_argentina()
        month = today.month
        year = today.year

    init_date = datetime(day=pid, month=month, year=year)
    duration_map = {"Mensual": 1, "Trimestral": 3, "Cuatrimestral": 4, "Semestral": 6, "Anual": 12}

    months = duration_map.get(boleta['plan_cobro'], 0)
    end_date = init_date + relativedelta(months=months)

    return init_date,end_date


def format_factura_afip(
    boleta: dict,
    merchant: dict,
    nro_comprobante: int,
    tipo_cbte: int,
    price: int,
    porcentaje_iva: float = 0,
    factura_existente: dict = None
) -> dict:

    factura = {}
    factura['FeCabReq'] = {
        'CantReg': 1,
        'PtoVta': merchant['afip']['pto_venta'],
        'CbteTipo': tipo_cbte
    }

    init_serv, end_serv = process_period(boleta)

    imp_iva = 0
    imp_neto = round(price, 2)
    imp_total = round(price, 2)

    if porcentaje_iva != 0:
        imp_neto = round(price / (1 + porcentaje_iva / 100), 2)
        imp_iva = round(imp_total - imp_neto, 2)

    print(f"Final Price: {price} = IVA: {imp_iva} + NETO: {imp_neto} (Tipo de IVA: {TIPO_IVA[porcentaje_iva]})")
    print('\n')

    tipo_doc = 'DNI'
    if tipo_cbte in [TIPO_COMPROBANTE.get("factura_a"), TIPO_COMPROBANTE.get("nota_credito_a")]:
        tipo_doc = 'CUIT'

    documento = boleta['cliente']['documento']
    if boleta['cliente']['factura_a']:
        documento = boleta['cliente']['cuit']

    fecha_factura = boleta["tries"].get("fecha_factura")
    if not fecha_factura:
        fecha_factura = boleta["tries"].get("date_approved")

    factura['FeDetReq'] = {
        'FECAEDetRequest': {
            'Concepto': TIPO_CONCEPTO['servicios'],
            'DocTipo': TIPO_DOCUMENTO.get(tipo_doc),
            'DocNro': documento,
            'CbteDesde': nro_comprobante,
            'CbteHasta': nro_comprobante,
            'CbteFch': fecha_factura.strftime("%Y%m%d"),
            'ImpTotal': imp_total,
            'ImpTotConc': 0,
            'ImpNeto': imp_neto,
            'ImpOpEx': 0,
            'ImpTrib': 0,
            'ImpIVA': imp_iva,
            'Iva': [
                {
                    'AlicIva':{
                        'Id': TIPO_IVA[porcentaje_iva],
                        'BaseImp': imp_neto,
                        'Importe': imp_iva
                    }
                }
            ],
            'FchServDesde': init_serv.strftime("%Y%m%d"),
            'FchServHasta': end_serv.strftime("%Y%m%d"),
            'FchVtoPago': fecha_factura.strftime("%Y%m%d"),
            'MonId': 'PES',
            'MonCotiz': 1,
        }
    }

    if tipo_cbte in [TIPO_COMPROBANTE.get("nota_credito_a"), TIPO_COMPROBANTE.get("nota_credito_b")]:
        factura['FeDetReq']['FECAEDetRequest']['CbtesAsoc'] = [{
            'CbteAsoc': {
                'Tipo': factura_existente['data_factura']['tipo_comprobante'],
                'PtoVta': factura_existente['data_factura']['pto_venta'],
                'Nro': factura_existente['data_factura']['num_comprobante'],
                'Cuit': factura_existente['data_facturante']['cuit'],
                'CbteFch': factura_existente['data_factura']['fecha_factura'].strftime("%Y%m%d")
            }
        }]

    return factura


def format_factura_db(
    boleta: dict,
    factura: dict,
    response_factura: dict,
    data_afip: dict
) -> dict:

    if 'payment_type' not in boleta['tries']:
        boleta['tries']['payment_type'] = '-'

    fecha_factura = boleta["tries"].get("fecha_factura")
    if not fecha_factura:
        fecha_factura = boleta["tries"].get("date_approved")

    factura_formateada = {
        'data_factura':{
            'tipo_comprobante': factura['FeCabReq']['CbteTipo'],
            'num_comprobante': response_factura['FeDetResp']['FECAEDetResponse'][0]['CbteDesde'],
            'pto_venta': factura['FeCabReq']['PtoVta'],
            'fecha_factura': fecha_factura,
            'mp_payment_id': boleta['tries']['payment_id'],
            'payment_type': boleta['tries']['payment_type'],
            'fecha_real_del_pago': boleta['tries'].get('date_approved', boleta['tries']['payment_day']),
            'source_boleta': boleta['source']
        },
        'data_facturante':{
            'cuit': data_afip['cuit'],
            'razon_social': data_afip['razon_social'],
            'inicio_de_actividades': data_afip['inicio_de_actividades'],
            'ii_bb': data_afip['ii_bb'],
            'domicilio': data_afip['domicilio'],
            'telefono': data_afip['telefono'],
            'email': data_afip['email'],
            'merchant_id': boleta['merchant_id']
        },
        'data_cliente':{
            'nombre': boleta['cliente']['nombre'],
            'apellido': boleta['cliente']['apellido'],
            'domicilio': boleta['cliente']['domicilio'],
            'email': boleta['cliente']['email'],
            'documento': factura['FeDetReq']['FECAEDetRequest']['DocNro']
        },
        'data_venta':{
            'inicio_servicios': datetime.strptime(factura['FeDetReq']['FECAEDetRequest']['FchServDesde'], '%Y%m%d'),
            'fin_servicios': datetime.strptime(factura['FeDetReq']['FECAEDetRequest']['FchServHasta'], '%Y%m%d'),
            'fecha_pago': boleta['tries'].get('date_approved', boleta['tries']['payment_day']),
            'importe_total': factura['FeDetReq']['FECAEDetRequest']['ImpTotal'],
            'importe_neto': factura['FeDetReq']['FECAEDetRequest']['ImpNeto'],
            'importe_iva': factura['FeDetReq']['FECAEDetRequest']['ImpIVA'],
        },
        'data_plan':{
            'plan_name': boleta['plan_name']
        },
        'data_cae': {
            'cae': response_factura['FeDetResp']['FECAEDetResponse'][0]['CAE'],
            'fecha_vencimiento': datetime.strptime(response_factura['FeDetResp']['FECAEDetResponse'][0]['CAEFchVto'], '%Y%m%d')
        },
        'boleta_id': boleta.get('_id')
    }
    if factura['FeCabReq']['CbteTipo'] in [TIPO_COMPROBANTE.get("nota_credito_a"), TIPO_COMPROBANTE.get("nota_credito_b")]:
        factura_formateada['data_nota_de_credito'] = {
            'factura_asociada':{
                'tipo_comprobante': factura['FeDetReq']['FECAEDetRequest']['CbtesAsoc'][0]['CbteAsoc']['Tipo'],
                'num_comprobante': factura['FeDetReq']['FECAEDetRequest']['CbtesAsoc'][0]['CbteAsoc']['Nro'],
                'pto_venta': factura['FeDetReq']['FECAEDetRequest']['CbtesAsoc'][0]['CbteAsoc']['PtoVta'],
                'cuit': factura['FeDetReq']['FECAEDetRequest']['CbtesAsoc'][0]['CbteAsoc']['Cuit'],
                'fecha_factura':  datetime.strptime(factura['FeDetReq']['FECAEDetRequest']['CbtesAsoc'][0]['CbteAsoc']['CbteFch'], '%Y%m%d')
            }
        }
    return factura_formateada


def realizar_facturacion_afip(
    nro_tipo_comprobante: int,
    price: float,
    boleta: dict,
    merchant: dict,
    afip: object,
    where_insert: str,
    porcentaje_iva: float,
    MAKE_UPDATE: bool,
    MAKE_PDF: bool,
    files_names: dict,
    factura_existente=None
):
    db = init_mongo()
    brand = db.brands.find_one({'_id': merchant['brand_id']})
    ultimo_comprobante = afip.ultimoAutorizado(nro_tipo_comprobante, merchant)
    nro_comprobante = ultimo_comprobante['CbteNro'] if ultimo_comprobante else 0
    print(f"Nro de ultimo comprobante: {nro_comprobante} para el tipo {nro_tipo_comprobante}")

    factura_afip = format_factura_afip(
        boleta=boleta,
        merchant=merchant,
        nro_comprobante=nro_comprobante+1,
        tipo_cbte=nro_tipo_comprobante,
        price=price,
        porcentaje_iva=porcentaje_iva,
        factura_existente=factura_existente
    )

    response_afip = afip.autorizarComprobante(factura_afip, merchant)
    response_afip = serialize_object(response_afip, dict)
    print('Response:')
    pprint(response_afip.get('FeCabResp'))
    print('\n')

    if not response_afip.get('FeCabResp'):
        return ["error", "ERROR: Fallo la autorizacion por parte de AFIP"]

    if response_afip.get('FeCabResp', {}).get('Resultado') != 'A':
        if observaciones := response_afip.get("FeDetResp", {}).get("FECAEDetResponse", {})[0].get("Observaciones"):
            print('Observaciones:')
            for obs in observaciones.get("Obs"):
                print(f"{obs.get('Code')}: {obs.get('Msg')}")
                print('\n')

        print('Errores:')
        for err in response_afip.get('Errors', {}).get('Err'):
            print(f"{err.get('Code')}: {err.get('Msg')}")
            print('\n')

        return ["error", "ERROR: Fallo la autorizacion por parte de AFIP"]

    factura_db = format_factura_db(
        boleta=boleta,
        factura=factura_afip,
        response_factura=response_afip,
        data_afip=merchant['afip']
    )

    factura_db['brand_id'] = brand['_id']
    if MAKE_UPDATE:
        print('Insertando la factura en la collection de facturas...')
        insert_factura = db.facturas.insert_one(factura_db)
        if not insert_factura.acknowledged:
            return ["error","error_insert_factura"]

        print(f"Se inserto una factura con el id {insert_factura.inserted_id}")
        data_update_boleta = {where_insert:insert_factura.inserted_id}
        print('Se inserta en la boleta:', data_update_boleta)
        result_update_boleta = db.boletas.update_one({'_id': boleta['_id']}, {'$set': data_update_boleta})
        print(f"Se encontro {result_update_boleta.matched_count} y se actualizo {result_update_boleta.modified_count}")

        if result_update_boleta.modified_count == 0:
            return ['error','error_update_boleta']

    pdf_filename = 'error_not_make_pdf'
    if MAKE_PDF:
        print('Generando pdf...')
        files_names['PDF_NAME'] = make_name_pdf(factura_db)

        pdf_filename = make_factura_pdf(
            files_names=files_names,
            factura=factura_db,
            brand=brand
        )
        print('PDF generado con exito')

    return ['successful', pdf_filename, insert_factura.inserted_id]



import setuptools
import subprocess

remote_version = (
    subprocess.run(["git", "describe", "--tags"], stdout=subprocess.PIPE)
    .stdout.decode("utf-8")
    .strip()
)
assert "." in remote_version

setuptools.setup(
    name="muvinai",
    version=remote_version,
    author="muvinai",
    description="Utilities for muvinai developers",
    long_description="Muvinai Python package",
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    package_data={"muvinai": ["VERSION"]},
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    install_requires=[
        'gspread',
        'cryptography',
        'python-dotenv',
        'mercadopago',
        'pymongo[srv]',
        'iso8601',
        'pyopenssl',
        'reportlab',
        'qrcode',
        'PyPDF2==2.12.1',
        'zeep',
        'xmltodict',
        'python-dateutil',
        'babel',
        'numpy',
        'muvisdk',
        'deepdiff'
    ],
)




from .boleta import Boleta
from .brand import Brand
from .card import Card
from .cliente import Cliente
from .corporativo import Corporativo
from .merchant import Merchant
from .muviBase import MuviBase
from .plan import Plan
from .sede import Sede



from typing import Optional
from muvisdk.SDK import SDK


class Card:
    def __init__(self, data: dict = None):
        self.data = data
        if not data:
            data = dict()
        self.id = data.get('id')
        self.mercadopago_id = data.get('mercadopago_id')
        self.decidir_id = data.get('decidir_id')
        self.card_type = data.get('card_type')
        self.card_brand = data.get('card_brand')
        self.first_six_digits = data.get('first_six_digits')
        self.payment_method_id = data.get('payment_method_id')

    def processors(self) -> list:
        processors = []
        if not self.data:
            return processors

        if 'decidir_id' in self.data:
            processors.append('decidir')
        if 'mercadopago_id' in self.data or 'id' in self.data:
            processors.append('mercadopago')

        return processors

    @classmethod
    def empty_card(cls):
        return Card()

    def get_id(self, processor: str) -> str:
        if processor == 'decidir':
            return self.decidir_id
        return self.mercadopago_id if self.mercadopago_id else self.id

    def tokenize(self, sdk: SDK) -> Optional[str]:
        response = sdk.card_token().create(self.data)
        if response["status"] > 299:
            print("         Falló la creación de token de tarjeta.")
            print(response["response"])
            return None

        return response["response"]["id"]



from bson import ObjectId

from muvinai.utilities.init_creds import init_mongo


class Brand:
    db = init_mongo()
    brands = None

    @classmethod
    def _get_brands(cls) -> dict:
        if not cls.brands:
            db = init_mongo()
            cls.brands = {b['_id']: b for b in db.brands.find({})}
        return cls.brands

    @classmethod
    def get(cls, _id: ObjectId = None, name: str = None):
        if not _id and not name:
            return
        db = init_mongo()
        if _id:
            brand_db = db.brands.find_one(_id)
            return Brand(brand_db) if brand_db else None

        brand_db = db.brands.find_one({'name': name})
        return Brand(brand_db) if brand_db else None

    @classmethod
    def search(cls, names: list = None) -> list:
        brands = cls._get_brands()
        if names:
            return [Brand(brand) for brand in brands.values() if brand['name'] in names]

        return [Brand(brand) for brand in brands.values()]

    def __init__(self, brand_db):
        self.brand_db = brand_db
        self.name = brand_db['name']
        self.id = brand_db['_id']

    def get_merchants(self) -> list:
        from .merchant import Merchant
        return [Merchant(merchant_db, brand=self) for merchant_db in self.db.merchants.find({'brand_id': self.id})]



from datetime import datetime
import pytz
from dateutil.relativedelta import *
import calendar
import typing

DIAS_GRACIA = 20
DIAS_ANTICIPO_DE_PAGO = 7
MID_DAY_HOUR = 12


arg_tz = pytz.timezone("America/Buenos_Aires")


def readable_date(f: datetime) -> str:
    """ Convierte una variable de tipo datetime.datetime al formato DD-MM-YYYY

    :param f: valor a convertir
    :type f: datetime
    :return: fecha formateada DD-MM-YYYY
    :rtype: str
    """
    return f.strftime("%d-%m-%Y")


def string_to_datetime(mongo_date: typing.Union[str, datetime]) -> datetime:
    """ Convierte un string a datetime

    :param mongo_date: Fecha a convertir
    :type mongo_date: str | datetime
    :return: Hora local
    :rtype: datetime
    """
    if type(mongo_date) == str:
        try:
            a = datetime.strptime(mongo_date, '%Y-%m-%dT%H:%M:%S.%f%z').astimezone(arg_tz)
        except:
            a = datetime.strptime(mongo_date, '%Y-%m-%dT%H:%M:%S%z').astimezone(arg_tz)
        return a
    else:
        mongo_date = arg_tz.localize(mongo_date + relativedelta(hours=-1))
        return mongo_date


def store_datetime_as_string(new_date: datetime) -> str:
    """ Convierte la fecha dada a string

    :param new_date: Fecha a convertir
    :type new_date: datetime.datetime
    :return: el valor recibido por parametro como string
    :rtype: str
    """
    a = new_date.astimezone(arg_tz)
    a = str(a).replace(" ", "T")  # todo    ESTA FORMULA TODAVÍA NO SIRVE.
    return a


def localize(utc_datetime: datetime) -> datetime:
    """ Convierte UTC a hora local

    :param utc_datetime: datetime en UTC+0
    :type utc_datetime: datetime
    :return: Hora recibida por parametro convertida a UTC+3 (hora local)
    :rtype: datetime
    """
    aware = utc_datetime.replace(tzinfo=pytz.utc)
    local_datetime = aware.astimezone(arg_tz)
    return local_datetime


def set_next_vigency(next_payment_date: datetime, cobro: str = 'Mensual') -> datetime:
    """ Establece la fecha de vigencia

    :param next_payment_date: Fecha del proximo pago
    :type next_payment_date: datetime
    :param cobro: Tipo de cobro (Mensual, Trimestral, Cuatrimestral, Semestral, Anual)
    :type cobro: str
    :return: Fecha de vigencia segun la fecha del proximo pago
    :rtype: datetime
    """
    days = DIAS_GRACIA + DIAS_ANTICIPO_DE_PAGO if cobro == 'Mensual' else DIAS_ANTICIPO_DE_PAGO
    return next_payment_date + relativedelta(days=days)


def get_periodo(source_date) -> str:
    """ Obtener el periodo a partir de una fecha de pago

    :param source_date:  datetime en UTC+0
    :type source_date: datetime
    :return: periodo actual con el formato MM/YYYY
    :rtype: str
    """
    month = source_date.month
    year = source_date.year
    return f"{month}/{year}"


def calculate_payment_date(pid: int, cobro: str, pub: str) -> datetime:
    """ Calcula el próximo día de pago para un cliente según la fecha de inicio de período.

    :param pid: Period Init Day - día del mes que inicia el periodo del usuario
    :type period_init_day: int
    :param cobro: los valores posibles son Mensual, Trimestral, Cuatrimestral, Semestral, Anual
    :type cobro: str
    :param pub: Periodo Última Boleta - el período de la última boleta existente para este usuario
    :type pub: str
    :return: Fecha de pago
    :rtype: datetime
    """
    month, year = [int(a) for a in pub.split('/')]
    last_period_date = datetime(year=year, month=month, day=1)
    duration_map = {"Mensual": 1, "Trimestral": 3, "Cuatrimestral": 4, "Semestral": 6, "Anual": 12}
    npd_aux = (last_period_date + relativedelta(months=duration_map[cobro]))
    npd = replace_day_in_period(npd_aux.year, npd_aux.month, pid)
    next_payment_date = (npd - relativedelta(days=DIAS_ANTICIPO_DE_PAGO)).replace(hour=15)
    return next_payment_date


def last_payment_date_adjust(client):
    day = client["period_init_day"]
    proposed = today_argentina() + relativedelta(days=DIAS_ANTICIPO_DE_PAGO)

    if proposed.day >= day:
        dif = proposed.day - day
    else:
        prev_month = proposed - relativedelta(months=1)
        month_days = calendar.monthrange(prev_month.year, prev_month.month)[1]
        dif = month_days - day + proposed.day
    return proposed - relativedelta(days=dif)


def today_argentina():
    return pytz.utc.localize(datetime.utcnow()).astimezone(arg_tz)


def replace_day_in_period(year: int, month: int, day: int):
    """ Calcula el próximo día de pago para un cliente según la fecha de inicio de período.

       :param year: año
       :type year: int
       :param month: mes
       :type month: int
       :param day: dia del mes que se pretende reemplazar en el periodo mes/año
       :type day: int
       :return: Fecha ajustada al periodo
       :rtype: datetime
       """
    try:
        date = datetime(year=year, month=month, day=day, hour=12)
    except ValueError:
        end_of_month = calendar.monthrange(year, month)[1]
        date = datetime(year=year, month=month, day=end_of_month, hour=12)
    return date


def get_periodo_anterior(cobro, periodo) -> str:
    """ Obtener el periodo a partir de una fecha de pago

    :param source_date:  datetime en UTC+0
    :type source_date: datetime
    :return: periodo actual con el formato MM/YYYY
    :rtype: str
    """
    month, year = [int(a) for a in periodo.split('/')]
    last_period_date = datetime(year=year, month=month, day=1)
    duration_map = {"Mensual": 1, "Trimestral": 3, "Cuatrimestral": 4, "Semestral": 6, "Anual": 12}
    aux = (last_period_date - relativedelta(months=duration_map[cobro]))
    month, year = aux.month, aux.year
    return f"{month}/{year}"


